<?php

namespace Webt\Services;

defined( '_JEXEC' ) or die;

use Joomla\CMS\Factory;
use Joomla\CMS\Log\Log;
use Joomla\CMS\Application\AdministratorApplication;
use Joomla\Component\Content\Site\Helper\AssociationHelper;
use Webt\Model\ArticleModel;
use Webt\Model\FieldModel;
use Webt\Model\SettingsModel;
use Webt\Model\AssociationModel;
use Exception;

/**
 * This is a PHP class named `TranslationManager`, responsible for managing article translations using translation services.
 *
 * @license     GNU General Public License version 2 or later, see License.txt
 */
class TranslationManager {

	/**
	 * @var $translationService Chosen translation service that will be used for translation
	 */
	public $translationService;
	/**
	 * Constructor for the `TranslationManager` class.
	 * Initializes the translation service based on the configured machine translation engine.
	 */
	public function __construct() {
		$settingsModel = new SettingsModel();
		if ( 'customProvider' === $settingsModel->getMtEngine() ) {
			$this->translationService = new CustomProviderService();
		} else {
			$this->translationService = new EtranslationService();
		}
	}
	/**
	 * Translates an article into multiple target languages.
	 *
	 * @param object $article The source article object to be translated.
	 * @param array  $targetLanguages An array of target language codes for translation.
	 * @param bool   $force_retranslate Flag to force retranslation even if the translation already exists.
	 * @param bool   $createAssociations Flag to create article associations between the source and translated articles.
	 * @param bool   $is_pretranslation Flag to indicate whether the translation is a pre-translation or not.
	 * @param array  $raw_translations Translation response from MT. If provided, does not call MT service.
	 * @return array An associative array containing the association keys and translated article IDs.
	 */
	public function translateArticle( $article, $targetLanguages, $force_retranslate = false, $createAssociations = true, $is_pretranslation = true, $raw_translations = null ) {
		$application = Factory::getApplication();

		$supportedLanguages = $this->filterSupportedLanguages( $targetLanguages );
		if ( count( $supportedLanguages ) < count( $targetLanguages ) && $application instanceof AdministratorApplication ) {
			$list_str = implode( ', ', array_diff( $targetLanguages, $supportedLanguages ) );
			$application->enqueueMessage( "Could not translate to languages not supported by MT provider: $list_str! If you wish to translate these languages, contact the MT provider support or choose a different MT provider.", 'error' );
		}

		$result         = array();
		$articleModel   = new ArticleModel();
		$sourceLanguage = $article->language;

		Log::add( "Starting article $article->id translation...", Log::DEBUG, 'webt' );
		foreach ( $supportedLanguages as $targetLanguage ) {
			$existingTranslation = $articleModel->getArticleTranslation( $article->id, $targetLanguage );
			if ( $existingTranslation && ! $force_retranslate ) {
				// skip translation, if it already exists.
				continue;
			}
			try {
				$translatableStrings    = $this->getTranslatableStrings( $article );
				$extractedTranslatables = TranslationHelper::extractTranslatableValues( $translatableStrings );

				if ( ! $raw_translations || empty( $raw_translations ) || count( $targetLanguages ) > 1 ) {
					$raw_translations = $this->translationService->translate( $sourceLanguage, $targetLanguage, $extractedTranslatables, $is_pretranslation, $article->id );
				}
				$translations = TranslationHelper::mergeTranslatedValues( $translatableStrings, $extractedTranslatables, $raw_translations );

				$translation_id = $this->translateArticleFromTextTranslation( $article->id, $targetLanguage, $translations, $createAssociations );
				if ( $translation_id ) {
					$result[ $article->id ][] = $translation_id;
				} else {
					throw new Exception( 'Translation could not be created!' );
				}
			} catch ( \Exception $e ) {
				$message     = (string) $e->getMessage();
				$log_message = "Could not translate article ($sourceLanguage -> $targetLanguage). Article ID: " . $article->id . ", error: '$message'";
				Log::add( $log_message, Log::ERROR, 'webt' );

				// Add error notification if translating from backend.
				if ( $application instanceof AdministratorApplication ) {
					$application->enqueueMessage( $log_message, 'error' );
				}
				continue;
			}
		}
		Log::add( "Finished article $article->id translation.", Log::DEBUG, 'webt' );

		return $result;
	}

	/**
	 * Retrieves all machine translatable strings for an article.
	 *
	 * @param object $article Article.
	 * @return string[] Translatable strings.
	 */
	public function getTranslatableStrings( $article ) {
		$strings = array();

		$this->traverseTranslatableValues(
			$article,
			function( $translatable_values, $model ) use ( &$strings ) {
				$strings = array_merge( $strings, $translatable_values );
			}
		);

		return $strings;
	}

	/**
	 * Group string translations by their origin (article or field)
	 *
	 * @param string[] $translations Translation list.
	 * @param object   $article Source article.
	 * @return array Multidimensional array of grouped translations.
	 */
	private function groupTranslations( $translations, $article ) {
		$i      = 0;
		$result = array(
			'article' => array(),
			'field'   => array(),
		);

		$this->traverseTranslatableValues(
			$article,
			function( $translatable_values, $model ) use ( $translations, &$result, &$i ) {
				$count              = count( $translatable_values );
				$slice              = array_values( array_slice( $translations, $i, $count ) );
				$named_translations = self::assignTranslationsToFieldNames( $model->translatableFields, $slice );
				$i                 += $count;

				if ( $model instanceof ArticleModel ) {
					$result['article'][] = $named_translations;
				} elseif ( $model instanceof FieldModel ) {
					$result['field'][] = $named_translations;
				}
			}
		);

		return $result;
	}

	/**
	 * Traverse all machine translatable strings for an article and act upon translatable value batches
	 *
	 * @param object   $article Article.
	 * @param callable $callback Callback function.
	 * @return void
	 */
	private function traverseTranslatableValues( $article, $callback ) {
		// Add article strings.
		$article_model   = new ArticleModel();
		$article_strings = $article_model->getTranslatableFieldValues( $article );
		$callback( $article_strings, $article_model );

		// Add field strings.
		$field_model = new FieldModel();
		$src_fields  = $field_model->getArticleFields( $article->id );

		$translatable_fields = array_filter(
			$src_fields,
			function( $src_field ) {
				return FieldModel::isFieldTranslatable( $src_field );
			}
		);

		foreach ( $translatable_fields as $field ) {
			$field_strings = $field_model->getTranslatableFieldValues( $field );
			$callback( $field_strings, $field_model );
		}
	}

	/**
	 * Updates or inserts fields for translated article.
	 *
	 * @param object $source_article Source article.
	 * @param object $translated_article_id Translated article ID.
	 * @return void
	 */
	private function upsertArticleFields( $source_article, $translated_article_id, $target_langcode, $translations ) {
		Log::add( "Setting fields for translated article $translated_article_id", Log::DEBUG, 'webt' );

		$field_model = new FieldModel();
		$src_fields  = $field_model->getArticleFields( $source_article->id );

		$translatable_fields = array_values(
			array_filter(
				$src_fields,
				function( $src_field ) {
					return FieldModel::isFieldTranslatable( $src_field );
				}
			)
		);

		$non_translatable_fields = array_values(
			array_filter(
				$src_fields,
				function( $src_field ) {
					return ! FieldModel::isFieldTranslatable( $src_field );
				}
			)
		);

		foreach ( $non_translatable_fields as $src_field ) {
			$field_model->insertFieldValue( $translated_article_id, $src_field->id, $src_field->value );
		}

		$translatable_field_count = count( $translatable_fields );
		for ( $i = 0; $i < $translatable_field_count; $i++ ) {
			$source_field = $translatable_fields[ $i ];
			$field_id     = $this->getTranslationFieldId( $source_field, $translations[ $i ], $target_langcode );

			$value = $translatable_fields[ $i ]->value;
			if ( $field_model->isFieldValueTranslatable( $source_field ) ) {
				$value = $translations[ $i ]['value'];
			} else {
				Log::add( "Field $source_field->id value is of non-translatable type '$source_field->type', setting original value", Log::DEBUG, 'webt' );
			}

			$field_model->insertFieldValue( $translated_article_id, $field_id, $value );
		}
	}

	/**
	 * Get translation field ID.
	 *
	 * @param array  $translatable_fields Source fields.
	 * @param string $translations field string translations.
	 * @param string $target_langcode Target language code.
	 * @return array Translation field ID.
	 */
	private function getTranslationFieldId( $source_field, $translations, $target_langcode ) {
		$field_model                = new FieldModel();
		$existing_field_translation = $field_model->getFieldTranslation( $source_field, $target_langcode );

		if ( $existing_field_translation ) {
			return $existing_field_translation->id;
		} else {
			return $field_model->insertFieldTranslation( $source_field, $translations, $target_langcode );
		}
	}

	/**
	 * Parses multidimensional translatable field array and assigns string translation values to field name keys.
	 *
	 * @param array $fields Translatable article fields.
	 * @param array $raw_translations Translations.
	 * @return array property => translation
	 */
	private static function assignTranslationsToFieldNames( $fields, $raw_translations ) {
		$fieldNameValues = array_values( $fields );
		$flattened_keys  = array();
		array_walk_recursive(
			$fieldNameValues,
			function( $a ) use ( &$flattened_keys ) {
				$flattened_keys[] = $a;
			}
		);
		return array_combine( $flattened_keys, $raw_translations );
	}

	/**
	 * Translates an article based on provided translations.
	 *
	 * @param int    $sourceArticleId The ID of the source article.
	 * @param string $targetLanguage The target language code for translation.
	 * @param array  $translations An array containing translations for the article's fields.
	 * @param bool   $createAssociations If false, skip association creation.
	 * @return int   Translated article ID.
	 */
	private function translateArticleFromTextTranslation( $sourceArticleId, $targetLanguage, $translations, $createAssociations = true ) {
		$articleModel     = new ArticleModel();
		$associationModel = new AssociationModel();

		$sourceArticle = $articleModel->getArticleById( $sourceArticleId );
		if ( $sourceArticle && $translations && ! empty( $translations ) ) {
			$grouped_translations = $this->groupTranslations( $translations, $sourceArticle );

			$existingTranslation = $articleModel->getArticleTranslation( $sourceArticleId, $targetLanguage );
			$id                  = $this->upsertTranslatedArticle( $sourceArticle, $targetLanguage, $grouped_translations['article'][0], $existingTranslation );

			$this->upsertArticleFields( $sourceArticle, $id, $targetLanguage, $grouped_translations['field'] );

			if ( $createAssociations ) {
				$assocation    = $associationModel->getArticleAssociation( $sourceArticleId );
				$assocationKey = $assocation ? $assocation->key : md5( json_encode( $translations ) );
				if ( ! $assocation ) {
					$associationModel->createArticleAssociation( $sourceArticleId, $assocationKey );
				}
				if ( ! $existingTranslation ) {
					$associationModel->createArticleAssociation( $id, $assocationKey );
				}
			}
			return $id;
		}
		return null;
	}

	/**
	 * Inserts or updates the translated article based on the translations received.
	 *
	 * @param object      $sourceArticle The source article object.
	 * @param string      $targetLanguage The target language code for translation.
	 * @param array       $translations An array containing translations for the article's fields.
	 * @param object|null $existingTranslation The existing translated article object, if available.
	 * @return int The ID of the inserted or updated translated article.
	 */
	private function upsertTranslatedArticle( $sourceArticle, $targetLanguage, $translations, $existingTranslation ) {
		$articleModel = new ArticleModel();
		// Replace language code in title.
		$translations['title'] = str_replace( $sourceArticle->language, $targetLanguage, $translations['title'] );

		// Create/update translation article.
		if ( $existingTranslation ) {
			$articleModel->updateArticle( $existingTranslation->id, $translations, $sourceArticle );
			return $existingTranslation->id;
		} else {
			$id = $articleModel->insertArticle( $this->getCategoryIdForLangcode( $sourceArticle, $targetLanguage ), $targetLanguage, $translations, $sourceArticle );
			return $id;
		}
	}
	/**
	 * Filters the supported target languages based on the available translation engines.
	 *
	 * @param array $targetLangcodes An array of target language codes for translation.
	 * @return array An array of supported target language codes.
	 */
	private function filterSupportedLanguages( $targetLangcodes ) {
		$engines            = $this->translationService->getSupportedEngineList();
		$settingsModel      = new SettingsModel();
		$defaultLangcode    = $settingsModel->getDefaultLanguage();
		$defaultLanguage    = explode( '-', $defaultLangcode )[0];
		$supportedLanguages = array();

		foreach ( $targetLangcodes as $targetLangcode ) {
			$targetLanguage = explode( '-', $targetLangcode )[0];
			foreach ( $engines as $engine ) {
				if ( $defaultLanguage === $engine->srcLang && $targetLanguage === $engine->trgLang ) {
					$supportedLanguages[] = $targetLangcode;
					break;
				}
			}
		}
		return array_unique( $supportedLanguages );
	}
	/**
	 * Gets the category ID for the target language based on category associations.
	 *
	 * @param object $sourceArticle The source article object.
	 * @param string $targetLanguage The target language code for translation.
	 * @return int The category ID for the target language, or the source article's category ID if not found.
	 */
	private function getCategoryIdForLangcode( $sourceArticle, $targetLanguage ) {
		$category_associations = AssociationHelper::getAssociations( $sourceArticle->catid, 'category' );
		if ( isset( $category_associations[ $targetLanguage ] ) ) {
			$params = explode( '&', $category_associations[ $targetLanguage ] );
			foreach ( $params as $param ) {
				if ( str_starts_with( $param, 'id=' ) ) {
					return explode( '=', $param )[1];
				}
			}
		}
		return $sourceArticle->catid;
	}

}
